@@ -57,6 +57,9 @@ module Text.XML.Stream.Parse
     , def
     , DecodeEntities
     , psDecodeEntities
+      -- *** Entity decoding
+    , decodeXmlEntities
+    , decodeHtmlEntities
       -- * Event parsing
     , tag
     , tagPredicate
@@ -81,7 +84,7 @@ module Text.XML.Stream.Parse
     ) where
 import Data.Attoparsec.Text
     ( char, Parser, takeWhile1, skipWhile, string
-    , manyTill, takeWhile, try, anyChar, endOfInput
+    , manyTill, takeWhile, try, anyChar
     )
 import qualified Control.Applicative as A
 import Data.Conduit.Attoparsec (sinkParser)
@@ -92,7 +95,8 @@ import Data.XML.Types
 
 import Filesystem.Path.CurrentOS (FilePath, encodeString)
 import Control.Applicative (Applicative(..), Alternative(empty,(<|>)), (<$>))
-import Data.Text (Text)
+import Data.Text (Text, pack)
+import Control.Arrow ((***))
 import qualified Data.Text as T
 import Data.Text.Read (Reader, decimal, hexadecimal)
 import Data.Text.Encoding (decodeUtf32BEWith)
@@ -115,8 +119,7 @@ import Control.Exception (Exception)
 import Data.Conduit.Binary (sourceFile)
 import Data.Char (isSpace)
 import Data.Default (Default (..))
-import Data.Maybe (catMaybes)
-import Control.Monad.Trans.Resource (ResourceIO, resourceThrow)
+import Control.Monad.Trans.Resource (MonadResource, monadThrow)
 import Control.Monad.Trans.Class (lift)
 
 type Ents = [(Text, Text)]
@@ -177,26 +180,24 @@ tnameToName _ (NSLevel _ m) (TName (Just pref) name) =
 -- first checks for BOMs, removing them as necessary, and then check for the
 -- equivalent of <?xml for each of UTF-8, UTF-16LE/BE, and UTF-32LE/BE. It
 -- defaults to assuming UTF-8.
-detectUtf :: C.ResourceThrow m => C.Conduit S.ByteString m TS.Text
+detectUtf :: C.MonadThrow m => C.Conduit S.ByteString m TS.Text
 detectUtf =
     conduit id
   where
-    conduit front = C.Conduit (push front) close
+    conduit front = C.NeedInput (push front) (return ())
 
-    push front bss = do
-        e <- getEncoding front bss
-        case e of
-            Left x -> return $ C.Producing (conduit x) []
-            Right (bss', decode) -> C.conduitPush decode bss'
-
-    close = return []
+    push front bss =
+        case getEncoding front bss of
+            Left x -> conduit x
+            Right (bss', C.NeedInput decode _) -> decode bss'
+            Right _ -> error "detectUtf: Unexpecting decode constructor"
 
     getEncoding front bs'
         | S.length bs < 4 =
-            return $ Left (bs `S.append`)
+            Left (bs `S.append`)
         | otherwise = do
             let decode = CT.decode codec
-            return $ Right (bsOut, decode)
+             in Right (bsOut, decode)
       where
         bs = front bs'
         bsOut = S.append (S.drop toDrop x) y
@@ -222,33 +223,32 @@ detectUtf =
 --
 -- This relies on 'detectUtf' to determine character encoding, and 'parseText'
 -- to do the actual parsing.
-parseBytes :: C.ResourceThrow m
+parseBytes :: C.MonadThrow m
            => ParseSettings -> C.Conduit S.ByteString m Event
 parseBytes ps = detectUtf C.=$= parseText ps
 
-dropBOM :: C.Resource m => C.Conduit TS.Text m TS.Text
-dropBOM = C.conduitState
-    False
-    push
-    close
+dropBOM :: Monad m => C.Conduit TS.Text m TS.Text
+dropBOM =
+    C.NeedInput push (return ())
   where
-    push a b = do
-        let (a', b') = go a b
-        return $ C.StateProducing a' [b']
-    close _ = return []
-    go True b = (True, b)
-    go False b =
-        case T.uncons b of
-            Nothing -> (False, b)
-            Just (c, cs) -> (True,
-                if c == '\xfeef' then cs else b)
+    push t =
+        case T.uncons t of
+            Nothing -> dropBOM
+            Just (c, cs) ->
+                let output
+                        | c == '\xfeef' = cs
+                        | otherwise = t
+                 in C.HaveOutput idConduit (return ()) output
+    idConduit = C.NeedInput
+        (\x -> C.HaveOutput idConduit (return ()) x)
+        (return ())
 
 -- | Parses a character stream into 'Event's. This function is implemented
 -- fully in Haskell using attoparsec-text for parsing. The produced error
 -- messages do not give line/column information, so you may prefer to stick
 -- with the parser provided by libxml-enumerator. However, this has the
 -- advantage of not relying on any C libraries.
-parseText :: C.ResourceThrow m
+parseText :: C.MonadThrow m
           => ParseSettings
           -> C.Conduit TS.Text m Event
 parseText de =
@@ -257,38 +257,27 @@ parseText de =
         C.=$= toEventC
         C.=$= addBeginEnd
   where
-    tokenize = C.sequenceSink () $ const $ C.Emit () . return <$> sinkToken de
-    addBeginEnd = C.conduitState
-        False
-        push
-        close
-      where
-        go False es = EventBeginDocument : [es]
-        go True es = [es]
+    tokenize = C.sequence $ sinkToken de
+    addBeginEnd = C.HaveOutput addEnd (return ()) EventBeginDocument
+    addEnd = C.NeedInput
+        (\e -> C.HaveOutput addEnd (return ()) e)
+        (C.HaveOutput (C.Done Nothing ()) (return ()) EventEndDocument)
 
-        push x es = return $ C.StateProducing True $ go x es
-        close x = return $ go x EventEndDocument
-
-toEventC :: C.Resource m => C.Conduit (Maybe Token) m Event
-toEventC = C.conduitState
-    ([], [])
-    push
-    close
+toEventC :: Monad m => C.Conduit Token m Event
+toEventC =
+    go [] []
   where
-    go es levels [] front = (es, levels, front [])
-    go es levels (t:ts) front =
-        let (es', levels', events) = tokenToEvent es levels t
-         in go es' levels' ts (front . (events ++))
-
-    push (es, levels) tokens =
-        return $ C.StateProducing (es', levels') events
+    go es levels =
+        C.NeedInput push close
       where
-        (es', levels', events) = go es levels (catMaybes [tokens]) id
+        push token =
+            send events $ go es' levels'
+          where
+            (es', levels', events) = tokenToEvent es levels token
+        close = C.Done Nothing ()
 
-    close (es, levels) =
-        return events
-      where
-        (_, _, events) = go es levels (catMaybes []) id
+    send [] c = c
+    send (e:es) c = C.HaveOutput (send es c) (return ()) e
 
 data ParseSettings = ParseSettings
     { psDecodeEntities :: DecodeEntities
@@ -296,11 +285,11 @@ data ParseSettings = ParseSettings
 
 instance Default ParseSettings where
     def = ParseSettings
-        { psDecodeEntities = decodeEntities
+        { psDecodeEntities = decodeXmlEntities
         }
 
-sinkToken :: C.ResourceThrow m => ParseSettings -> C.Sink TS.Text m (Maybe Token)
-sinkToken de = sinkParser ((endOfInput >> return Nothing) <|> fmap Just (parseToken $ psDecodeEntities de))
+sinkToken :: C.MonadThrow m => ParseSettings -> C.Sink TS.Text m Token
+sinkToken = sinkParser . parseToken . psDecodeEntities
 
 parseToken :: DecodeEntities -> Parser Token
 parseToken de = (char '<' >> parseLt) <|> TokenContent <$> parseContent de False False
@@ -472,13 +461,13 @@ data ContentType =
 -- | Grabs the next piece of content if available. This function skips over any
 -- comments and instructions and concatenates all content until the next start
 -- or end tag.
-contentMaybe :: C.ResourceThrow m => C.Sink Event m (Maybe Text)
+contentMaybe :: C.MonadThrow m => C.Sink Event m (Maybe Text)
 contentMaybe = do
     x <- CL.peek
     case pc' x of
         Ignore -> CL.drop 1 >> contentMaybe
         IsContent t -> CL.drop 1 >> fmap Just (takeContents (t:))
-        IsError e -> lift $ C.resourceThrow $ XmlException e x
+        IsError e -> lift $ C.monadThrow $ XmlException e x
         NotContent -> return Nothing
   where
     pc' Nothing = NotContent
@@ -499,12 +488,12 @@ contentMaybe = do
         case pc' x of
             Ignore -> CL.drop 1 >> takeContents front
             IsContent t -> CL.drop 1 >> takeContents (front . (:) t)
-            IsError e -> lift $ C.resourceThrow $ XmlException e x
+            IsError e -> lift $ C.monadThrow $ XmlException e x
             NotContent -> return $ T.concat $ front []
 
 -- | Grabs the next piece of content. If none if available, returns 'T.empty'.
 -- This is simply a wrapper around 'contentMaybe'.
-content :: C.ResourceThrow m => C.Sink Event m Text
+content :: C.MonadThrow m => C.Sink Event m Text
 content = do
     x <- contentMaybe
     case x of
@@ -520,7 +509,7 @@ content = do
 -- consumed. If you want to allow extra attributes, see 'ignoreAttrs'.
 --
 -- This function automatically ignores comments, instructions and whitespace.
-tag :: C.ResourceThrow m
+tag :: C.MonadThrow m
     => (Name -> Maybe a)
     -> (a -> AttrParser b)
     -> (b -> C.Sink Event m c)
@@ -532,7 +521,7 @@ tag checkName attrParser f = do
             case checkName name of
                 Just y ->
                     case runAttrParser' (attrParser y) as of
-                        Left e -> lift $ C.resourceThrow e
+                        Left e -> lift $ C.monadThrow e
                         Right z -> do
                             CL.drop 1
                             z' <- f z
@@ -540,7 +529,7 @@ tag checkName attrParser f = do
                             case a of
                                 Just (EventEndElement name')
                                     | name == name' -> CL.drop 1 >> return (Just z')
-                                _ -> lift $ C.resourceThrow $ XmlException ("Expected end tag for: " ++ show name) a
+                                _ -> lift $ C.monadThrow $ XmlException ("Expected end tag for: " ++ show name) a
                 Nothing -> return Nothing
         _ -> return Nothing
   where
@@ -570,7 +559,7 @@ tag checkName attrParser f = do
             Right (attr, _) -> Left $ UnparsedAttributes attr
 
 -- | A simplified version of 'tag' which matches against boolean predicates.
-tagPredicate :: C.ResourceThrow m
+tagPredicate :: C.MonadThrow m
              => (Name -> Bool)
              -> AttrParser a
              -> (a -> C.Sink Event m b)
@@ -581,7 +570,7 @@ tagPredicate p attrParser = tag (\x -> if p x then Just () else Nothing) (const
 -- of taking a predicate function. This is often sufficient, and when combined
 -- with OverloadedStrings and the IsString instance of 'Name', can prove to be
 -- very concise.
-tagName :: C.ResourceThrow m
+tagName :: C.MonadThrow m
      => Name
      -> AttrParser a
      -> (a -> C.Sink Event m b)
@@ -589,14 +578,14 @@ tagName :: C.ResourceThrow m
 tagName name = tagPredicate (== name)
 
 -- | A further simplified tag parser, which requires that no attributes exist.
-tagNoAttr :: C.ResourceThrow m => Name -> C.Sink Event m a -> C.Sink Event m (Maybe a)
+tagNoAttr :: C.MonadThrow m => Name -> C.Sink Event m a -> C.Sink Event m (Maybe a)
 tagNoAttr name f = tagName name (return ()) $ const f
 
 -- | Get the value of the first parser which returns 'Just'. If no parsers
 -- succeed (i.e., return 'Just'), this function returns 'Nothing'.
 --
 -- > orE a b = choose [a, b]
-orE :: C.Resource m => C.Sink Event m (Maybe a) -> C.Sink Event m (Maybe a) -> C.Sink Event m (Maybe a)
+orE :: Monad m => C.Sink Event m (Maybe a) -> C.Sink Event m (Maybe a) -> C.Sink Event m (Maybe a)
 orE a b = do
   x <- a
   case x of
@@ -605,7 +594,7 @@ orE a b = do
 
 -- | Get the value of the first parser which returns 'Just'. If no parsers
 -- succeed (i.e., return 'Just'), this function returns 'Nothing'.
-choose :: C.Resource m
+choose :: Monad m
        => [C.Sink Event m (Maybe a)]
        -> C.Sink Event m (Maybe a)
 choose [] = return Nothing
@@ -618,27 +607,27 @@ choose (i:is) = do
 -- | Force an optional parser into a required parser. All of the 'tag'
 -- functions, 'choose' and 'many' deal with 'Maybe' parsers. Use this when you
 -- want to finally force something to happen.
-force :: C.ResourceThrow m
+force :: C.MonadThrow m
       => String -- ^ Error message
       -> C.Sink Event m (Maybe a)
       -> C.Sink Event m a
 force msg i = do
     x <- i
     case x of
-        Nothing -> lift $ resourceThrow $ XmlException msg Nothing
+        Nothing -> lift $ monadThrow $ XmlException msg Nothing
         Just a -> return a
 
 -- | A helper function which reads a file from disk using 'enumFile', detects
 -- character encoding using 'detectUtf', parses the XML using 'parseBytes', and
 -- then hands off control to your supplied parser.
-parseFile :: (ResourceIO m, C.ResourceThrow m)
+parseFile :: MonadResource m
           => ParseSettings
           -> FilePath
           -> C.Source m Event
 parseFile ps fp = sourceFile (encodeString fp) C.$= parseBytes ps
 
 -- | Parse an event stream from a lazy 'L.ByteString'.
-parseLBS :: C.ResourceThrow m
+parseLBS :: C.MonadThrow m
          => ParseSettings
          -> L.ByteString
          -> C.Source m Event
@@ -724,7 +713,7 @@ ignoreAttrs :: AttrParser ()
 ignoreAttrs = AttrParser $ \_ -> Right ([], ())
 
 -- | Keep parsing elements as long as the parser returns 'Just'.
-many :: C.Resource m => C.Sink Event m (Maybe a) -> C.Sink Event m [a]
+many :: Monad m => C.Sink Event m (Maybe a) -> C.Sink Event m [a]
 many i =
     go id
   where
@@ -738,13 +727,13 @@ type DecodeEntities = Text -> Content
 
 -- | Default implementation of 'DecodeEntities': handles numeric entities and
 -- the five standard character entities (lt, gt, amp, quot, apos).
-decodeEntities :: DecodeEntities
-decodeEntities "lt" = ContentText "<"
-decodeEntities "gt" = ContentText ">"
-decodeEntities "amp" = ContentText "&"
-decodeEntities "quot" = ContentText "\""
-decodeEntities "apos" = ContentText "'"
-decodeEntities t = let backup = ContentEntity t in
+decodeXmlEntities :: DecodeEntities
+decodeXmlEntities "lt" = ContentText "<"
+decodeXmlEntities "gt" = ContentText ">"
+decodeXmlEntities "amp" = ContentText "&"
+decodeXmlEntities "quot" = ContentText "\""
+decodeXmlEntities "apos" = ContentText "'"
+decodeXmlEntities t = let backup = ContentEntity t in
   case T.uncons t of
     Just ('#', t') ->
       case T.uncons t' of
@@ -756,6 +745,274 @@ decodeEntities t = let backup = ContentEntity t in
           | otherwise        -> decodeChar decimal backup t'
     _ -> backup
 
+-- | HTML4-compliant entity decoder. Handles numerics, the five standard
+-- character entities, and the additional 248 entities defined by HTML 4 and
+-- XHTML 1.
+--
+-- Note that HTML 5 introduces a drastically larger number of entities, and
+-- this code does not recognize most of them.
+decodeHtmlEntities :: DecodeEntities
+decodeHtmlEntities t =
+    case decodeXmlEntities t of
+        x@ContentText{} -> x
+        backup@ContentEntity{} ->
+            case Map.lookup t htmlEntities of
+                Just x -> ContentText x
+                Nothing -> backup
+
+htmlEntities :: Map.Map T.Text T.Text
+htmlEntities = Map.fromList
+    $ map (pack *** pack) -- Work around the long-compile-time bug
+    [ ("nbsp", "\160")
+    , ("iexcl", "\161")
+    , ("cent", "\162")
+    , ("pound", "\163")
+    , ("curren", "\164")
+    , ("yen", "\165")
+    , ("brvbar", "\166")
+    , ("sect", "\167")
+    , ("uml", "\168")
+    , ("copy", "\169")
+    , ("ordf", "\170")
+    , ("laquo", "\171")
+    , ("not", "\172")
+    , ("shy", "\173")
+    , ("reg", "\174")
+    , ("macr", "\175")
+    , ("deg", "\176")
+    , ("plusmn", "\177")
+    , ("sup2", "\178")
+    , ("sup3", "\179")
+    , ("acute", "\180")
+    , ("micro", "\181")
+    , ("para", "\182")
+    , ("middot", "\183")
+    , ("cedil", "\184")
+    , ("sup1", "\185")
+    , ("ordm", "\186")
+    , ("raquo", "\187")
+    , ("frac14", "\188")
+    , ("frac12", "\189")
+    , ("frac34", "\190")
+    , ("iquest", "\191")
+    , ("Agrave", "\192")
+    , ("Aacute", "\193")
+    , ("Acirc", "\194")
+    , ("Atilde", "\195")
+    , ("Auml", "\196")
+    , ("Aring", "\197")
+    , ("AElig", "\198")
+    , ("Ccedil", "\199")
+    , ("Egrave", "\200")
+    , ("Eacute", "\201")
+    , ("Ecirc", "\202")
+    , ("Euml", "\203")
+    , ("Igrave", "\204")
+    , ("Iacute", "\205")
+    , ("Icirc", "\206")
+    , ("Iuml", "\207")
+    , ("ETH", "\208")
+    , ("Ntilde", "\209")
+    , ("Ograve", "\210")
+    , ("Oacute", "\211")
+    , ("Ocirc", "\212")
+    , ("Otilde", "\213")
+    , ("Ouml", "\214")
+    , ("times", "\215")
+    , ("Oslash", "\216")
+    , ("Ugrave", "\217")
+    , ("Uacute", "\218")
+    , ("Ucirc", "\219")
+    , ("Uuml", "\220")
+    , ("Yacute", "\221")
+    , ("THORN", "\222")
+    , ("szlig", "\223")
+    , ("agrave", "\224")
+    , ("aacute", "\225")
+    , ("acirc", "\226")
+    , ("atilde", "\227")
+    , ("auml", "\228")
+    , ("aring", "\229")
+    , ("aelig", "\230")
+    , ("ccedil", "\231")
+    , ("egrave", "\232")
+    , ("eacute", "\233")
+    , ("ecirc", "\234")
+    , ("euml", "\235")
+    , ("igrave", "\236")
+    , ("iacute", "\237")
+    , ("icirc", "\238")
+    , ("iuml", "\239")
+    , ("eth", "\240")
+    , ("ntilde", "\241")
+    , ("ograve", "\242")
+    , ("oacute", "\243")
+    , ("ocirc", "\244")
+    , ("otilde", "\245")
+    , ("ouml", "\246")
+    , ("divide", "\247")
+    , ("oslash", "\248")
+    , ("ugrave", "\249")
+    , ("uacute", "\250")
+    , ("ucirc", "\251")
+    , ("uuml", "\252")
+    , ("yacute", "\253")
+    , ("thorn", "\254")
+    , ("yuml", "\255")
+    , ("OElig", "\338")
+    , ("oelig", "\339")
+    , ("Scaron", "\352")
+    , ("scaron", "\353")
+    , ("Yuml", "\376")
+    , ("fnof", "\402")
+    , ("circ", "\710")
+    , ("tilde", "\732")
+    , ("Alpha", "\913")
+    , ("Beta", "\914")
+    , ("Gamma", "\915")
+    , ("Delta", "\916")
+    , ("Epsilon", "\917")
+    , ("Zeta", "\918")
+    , ("Eta", "\919")
+    , ("Theta", "\920")
+    , ("Iota", "\921")
+    , ("Kappa", "\922")
+    , ("Lambda", "\923")
+    , ("Mu", "\924")
+    , ("Nu", "\925")
+    , ("Xi", "\926")
+    , ("Omicron", "\927")
+    , ("Pi", "\928")
+    , ("Rho", "\929")
+    , ("Sigma", "\931")
+    , ("Tau", "\932")
+    , ("Upsilon", "\933")
+    , ("Phi", "\934")
+    , ("Chi", "\935")
+    , ("Psi", "\936")
+    , ("Omega", "\937")
+    , ("alpha", "\945")
+    , ("beta", "\946")
+    , ("gamma", "\947")
+    , ("delta", "\948")
+    , ("epsilon", "\949")
+    , ("zeta", "\950")
+    , ("eta", "\951")
+    , ("theta", "\952")
+    , ("iota", "\953")
+    , ("kappa", "\954")
+    , ("lambda", "\955")
+    , ("mu", "\956")
+    , ("nu", "\957")
+    , ("xi", "\958")
+    , ("omicron", "\959")
+    , ("pi", "\960")
+    , ("rho", "\961")
+    , ("sigmaf", "\962")
+    , ("sigma", "\963")
+    , ("tau", "\964")
+    , ("upsilon", "\965")
+    , ("phi", "\966")
+    , ("chi", "\967")
+    , ("psi", "\968")
+    , ("omega", "\969")
+    , ("thetasym", "\977")
+    , ("upsih", "\978")
+    , ("piv", "\982")
+    , ("ensp", "\8194")
+    , ("emsp", "\8195")
+    , ("thinsp", "\8201")
+    , ("zwnj", "\8204")
+    , ("zwj", "\8205")
+    , ("lrm", "\8206")
+    , ("rlm", "\8207")
+    , ("ndash", "\8211")
+    , ("mdash", "\8212")
+    , ("lsquo", "\8216")
+    , ("rsquo", "\8217")
+    , ("sbquo", "\8218")
+    , ("ldquo", "\8220")
+    , ("rdquo", "\8221")
+    , ("bdquo", "\8222")
+    , ("dagger", "\8224")
+    , ("Dagger", "\8225")
+    , ("bull", "\8226")
+    , ("hellip", "\8230")
+    , ("permil", "\8240")
+    , ("prime", "\8242")
+    , ("Prime", "\8243")
+    , ("lsaquo", "\8249")
+    , ("rsaquo", "\8250")
+    , ("oline", "\8254")
+    , ("frasl", "\8260")
+    , ("euro", "\8364")
+    , ("image", "\8465")
+    , ("weierp", "\8472")
+    , ("real", "\8476")
+    , ("trade", "\8482")
+    , ("alefsym", "\8501")
+    , ("larr", "\8592")
+    , ("uarr", "\8593")
+    , ("rarr", "\8594")
+    , ("darr", "\8595")
+    , ("harr", "\8596")
+    , ("crarr", "\8629")
+    , ("lArr", "\8656")
+    , ("uArr", "\8657")
+    , ("rArr", "\8658")
+    , ("dArr", "\8659")
+    , ("hArr", "\8660")
+    , ("forall", "\8704")
+    , ("part", "\8706")
+    , ("exist", "\8707")
+    , ("empty", "\8709")
+    , ("nabla", "\8711")
+    , ("isin", "\8712")
+    , ("notin", "\8713")
+    , ("ni", "\8715")
+    , ("prod", "\8719")
+    , ("sum", "\8721")
+    , ("minus", "\8722")
+    , ("lowast", "\8727")
+    , ("radic", "\8730")
+    , ("prop", "\8733")
+    , ("infin", "\8734")
+    , ("ang", "\8736")
+    , ("and", "\8743")
+    , ("or", "\8744")
+    , ("cap", "\8745")
+    , ("cup", "\8746")
+    , ("int", "\8747")
+    , ("there4", "\8756")
+    , ("sim", "\8764")
+    , ("cong", "\8773")
+    , ("asymp", "\8776")
+    , ("ne", "\8800")
+    , ("equiv", "\8801")
+    , ("le", "\8804")
+    , ("ge", "\8805")
+    , ("sub", "\8834")
+    , ("sup", "\8835")
+    , ("nsub", "\8836")
+    , ("sube", "\8838")
+    , ("supe", "\8839")
+    , ("oplus", "\8853")
+    , ("otimes", "\8855")
+    , ("perp", "\8869")
+    , ("sdot", "\8901")
+    , ("lceil", "\8968")
+    , ("rceil", "\8969")
+    , ("lfloor", "\8970")
+    , ("rfloor", "\8971")
+    , ("lang", "\9001")
+    , ("rang", "\9002")
+    , ("loz", "\9674")
+    , ("spades", "\9824")
+    , ("clubs", "\9827")
+    , ("hearts", "\9829")
+    , ("diams", "\9830")
+    ]
+
 decodeChar :: Reader Word32 -> Content -> Text -> Content
 decodeChar readNum backup = either (const backup) toContent . readNum
   where
